/*
 * Copyright (c) 2015 by inovat, innovative systeme - verkehr - tunnel - technik, Dipl.-Ing. H. C. Kniss
 *
 * This file is part of de.bsvrz.kex.tls.osi7.Eak
 *
 * de.bsvrz.kex.tls.osi7.Eak is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  de.bsvrz.kex.tls.osi7.Eak is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with de.bsvrz.kex.tls.osi7.Eak.  If not, see <http://www.gnu.org/licenses/>.
 *
 *  Diese Datei ist Teil von de.bsvrz.kex.tls.osi7.Eak.
 *
 *  de.bsvrz.kex.tls.osi7.Eak ist Freie Software: Sie knnen es unter den Bedingungen
 *  der GNU General Public License, wie von der Free Software Foundation,
 *  Version 3 der Lizenz oder (nach Ihrer Wahl) jeder spteren
 *  verffentlichten Version, weiterverbreiten und/oder modifizieren.
 *
 *  de.bsvrz.kex.tls.osi7.Eak wird in der Hoffnung, dass es ntzlich sein wird, aber
 *  OHNE JEDE GEWHRLEISTUNG, bereitgestellt; sogar ohne die implizite
 *  Gewhrleistung der MARKTFHIGKEIT oder EIGNUNG FR EINEN BESTIMMTEN ZWECK.
 *  Siehe die GNU General Public License fr weitere Details.
 *
 *  Sie sollten eine Kopie der GNU General Public License zusammen mit diesem
 *  Programm erhalten haben. Wenn nicht, siehe <http://www.gnu.org/licenses/>.
 *
 * Contact Information:
 * inovat, Dipl.-Ing. H. C. Kniss
 * An der Krautwiese 37
 * D-53783 Eitorf
 * +49 (0)2243 8464 193
 * info@inovat.de
 * www.inovat.de
 */

package de.bsvrz.kex.tls.osi7;

import de.bsvrz.dav.daf.main.*;
import de.bsvrz.dav.daf.main.config.*;
import de.bsvrz.kex.tls.osi7.conversion.EakConversionInterface;
import de.bsvrz.sys.funclib.debug.Debug;

import java.lang.reflect.InvocationTargetException;
import java.util.*;
import java.util.regex.Pattern;

/**
 * TBD: Kurzbeschreibung TBD: Langebeschreibung
 *
 * @author inovat, innovative systeme - verkehr - tunnel - technik
 * @author Dipl.-Ing. Hans Christian Kni (HCK)
 * @version $Revision$ / $Date$ / ($Author$)
 */
public class Eak implements ClientReceiverInterface {
	// ------------------------------------------------------------------------
	// VARIABLENDEKLARATIONEN
	// ------------------------------------------------------------------------

	// Sonstige ---------------------------------------------------------------
	/** DebugLogger fr Debug-Ausgaben */
	private static final Debug debug = Debug.getLogger();

	/**
	 * Hashtabelle aller De an einem Eak. Die Identifikation erfolgt ber einen Schlssel auf "FG-DeNummer", z.B. "4-193".
	 * Fr jedes De wird  diese FG-De Kombinationen eingetragen. Dadurch lt sich bei der Verarbeitung der De-Blcke das
	 * fr die Verarbeitung zustndige De einfach aus der FG des Einzeltelegramms und einer DE-Nummer (ungleich 255!) des
	 * enthaltenen DE-Blocks ermitteln.
	 */
	Map _deHashTable = new Hashtable();

	/** Liste der angeschlossenen De. */
	List<De> _deList = new ArrayList<De>();
	// Private ----------------------------------------------------------------
	/** Klassenobjekt des UmsetzungsModuls fr die Konvertierung der Telegramme in Attributgruppen (und zurck) */
	private Class _conversionClass;

	/** DaV Interface */
	private final ClientDavInterface _daf;

	/** Objektreferenz auf das Objekt vom "typ.eak" aus dem Teilmodell TLS. */
	private final ConfigurationObject _eak;

	/** Instanz des UmsetzungsModuls fr die Konvertierung der Telegramme in Attributgruppen (und zurck) */
	private EakConversionInterface _conversionModul;

	/**
	 * Liste der zwischengespeicherten DE-Blcke fr dieses EAK, die erst nach externer Triggerung zu einem/mehreren
	 * Einzeltelegramm zusammengestellt werden.
	 */
	private List<DeBlockToSend> _bufferedDeBlocks = new ArrayList<DeBlockToSend>();

	/** Objektreferenz auf das bergeordnete Sm-Objekt. */
	private final Sm _sm;

	/** Schlusselarray auf Basis der Funktionsgruppe und der angeschlossenen DE als String. */
	private String[] _eakKeys;

	/** Name des UmsetzungsModul fr die Konvertierung der Telegramme in Attributgruppen (und zurck) */
	private String _conversionModulName;

	/** Bezeichnung des Gertes */
	private final String _eakDescription;

	/** OSI2 KnotenNummer */
	private final int _osi2Adress;

	// ------------------------------------------------------------------------
	// KONSTRUKTOREN  (und vom Konstruktor verwende Methoden)
	// ------------------------------------------------------------------------

	/** Erzeugt eine Instanz eines Eak. */
	public Eak(ConfigurationObject eak, ClientDavInterface daf, Sm sm) {
		_eak = eak;
		_daf = daf;
		_sm = sm;

		DataModel config = _daf.getDataModel();
		AttributeGroup eakPropertiesAtg = config.getAttributeGroup("atg.eak");
		Data eakProperties = _eak.getConfigurationData(eakPropertiesAtg);
		_eakDescription = eakProperties.getTextValue("Bezeichnung").getText();
		_osi2Adress = eakProperties.getScaledValue("OSI2Adresse").intValue();
		_conversionModulName = eakProperties.getTextValue("UmsetzungsModul").getText().trim();

		// Wenn UmsetzungsModul nicht speziell konfiguriert wurde, dann abhngig
		// von der FG de DE ein passendes StandardModul vorbelegen.
		if (_conversionModulName.equals("")) {
			_conversionModulName = "de.bsvrz.kex.tls.osi7.conversion.EakDefault";
		}

		installConversionModule();

		// Liste der angeschlossenen DE ermitteln.
		List deList = _eak.getObjectSet("De").getElements();
		Iterator deIterator = deList.iterator();
		De de;
		_eakKeys = new String[deList.size()];
		int i = 0;
		while (deIterator.hasNext()) {
			de = new De((ConfigurationObject) (deIterator.next()), daf, this);
			// Liste der De aufbauen
			_deList.add(de);
			// De jetzt mit seinem FG-DE Schlssel in die Tabelle mit den De eintragen.
			_deHashTable.put(de.getKey(), de);
			// FG-DE Schlssel in die Tabelle mit den _eakKeys eintragen.
			_eakKeys[i++] = de.getKey();
		}

		// Triggerdatenart fr Versand gepufferter DE anmelden
		subcribeTrigger();

		//COM: debug.fine(toString());
	}

	private void installConversionModule() {
		try {
			// Anpassung zur Kompatibilitt nach Umstellung auf neue Packagestruktur de.bsvrz.*
			// Liegt Umsetzungsklasse noch im alten Package kex.tls.osi7.conversion.*, dann
			// wird automatisch im Package de.bsvrz.kex.tls.osi7.conversion.* gesucht.
			debug.fine("KonvertierungsModulName Vorgabe: [" + _conversionModulName + "]");
			if (Pattern.matches("^kex\\.tls\\.osi7\\.conversion\\..*", _conversionModulName)) {
				_conversionModulName = ("de.bsvrz." + _conversionModulName).trim();

			}
			debug.fine("KonvertierungsModulName nach Anpassung auf neue Packagestruktur de.bsvrz.*: [" + _conversionModulName + "]");

			// Umsetungsklasse laden und instanzieren.
			debug.finest("Versuche Umsetzungsklasse zu laden: " + _conversionModulName);
			_conversionClass = Class.forName(_conversionModulName);
			_conversionModul = (EakConversionInterface) _conversionClass.getConstructor(new Class[]{ClientDavInterface.class, Sm.class, Eak.class}).newInstance(new Object[]{_daf, getSm(), this});
		}
		catch (ClassNotFoundException e) {
			debug.error("Konnte UmsetzungsModul [" + _conversionModulName + "] nicht laden!", e);
		}
		catch (NoSuchMethodException e) {
			debug.error("Konstruktoraufruf in Klasse[" + _conversionModulName + "] fehlgeschlagen!", e);
		}
		catch (IllegalAccessException e) {
			debug.error("Zugriff auf UmsetzungsModul [" + _conversionModulName + "] fehlgeschlagen!", e);
		}
		catch (InstantiationException e) {
			debug.error("UmsetzungsModul [" + _conversionModulName + "] konnte  nicht instanziert werden!", e);
		}
		catch (InvocationTargetException e) {
			debug.error("Konstruktoraufruf am UmsetzungsModul [" + _conversionModulName + "] fehlgeschlagen!", e);
		}
	}

	/** Triggerdatenart fr Versand gepufferter DE anmelden */
	private void subcribeTrigger() {
		DataModel config = _daf.getDataModel();
		AttributeGroup atgSendSingleTelegram = config.getAttributeGroup("atg.tlsEinzelTelegrammSenden");
		Aspect aspect = config.getAspect("asp.tlsVorgabe");
		DataDescription dataDescription = new DataDescription(atgSendSingleTelegram, aspect);
		_daf.subscribeReceiver(this, _eak, dataDescription, new ReceiveOptions(false), new ReceiverRole(true));// Als Senke anmelden !!
	}

	/**
	 * Zustzlichen Konstruktor fr Eak,  die nicht ber die TLS-Versorgung sondern ber die Tabelle der zustzlichen
	 * Kommunikationspartner versorgt worden sind.
	 */
	public Eak(ClientDavInterface daf, Sm sm, List additionalDeviceTable, Map conversionModules) {
		//noinspection AssignmentToNull
		_eak = null;
		_daf = daf;
		_sm = sm;

		_eakDescription = "Virtuelles EAK fr TLS hnliche Kommunikation";
		_osi2Adress = -1;// virtuelles EAK, hat keine (gltige) OSI-2-Adresse.

		DataModel config = _daf.getDataModel();

		try {
			_conversionModulName = ((String) conversionModules.get(config.getObject("typ.eak"))).trim();
		}
		catch (NullPointerException e) {
			debug.error("UmsetzungsModul fr EAK (bei TLS hnlicher Kommunikation) muss \n" + "in [atg.gertTlshnlicheKommunikationsPartnerUmsetzungsModule] \n" + "einmal fr den Typ [typ.eak] fr das entsprechende Gert (VIZ, VRZ, UZ) angegeben werden!", e);
		}

		installConversionModule();

		// Liste der angeschlossenen TLS hnlichen Konfigurationspartner ("DE") ermitteln.
		List deList = new ArrayList();

		ListIterator iterator = additionalDeviceTable.listIterator();
		while (iterator.hasNext()) {
			AdditionalCommunicationPartner acp = (AdditionalCommunicationPartner) iterator.next();

			// Wenn zustzliches "DE" an Knoten des bergeordneten SM gehrt und noch nicht "gelscht" wurde, dann...
			if ((acp.getKey().equals(_sm.getKey())) && (!acp.isDeleted())) {
				deList.add(acp);// ...Daten merken...
				acp.setDeleted(true);//...und aus der Gesamtliste der zustzlichen KommunikationsPartner "lschen".
			}
		}
		Iterator deIterator = deList.iterator();
		De de;
		_eakKeys = new String[deList.size()];
		int i = 0;
		while (deIterator.hasNext()) {
			de = new De((AdditionalCommunicationPartner) (deIterator.next()), daf, this, conversionModules);
			// Liste der De aufbauen
			_deList.add(de);
			// De jetzt mit seinem FG-DE Schlssel in die Tabelle mit den De eintragen.
			_deHashTable.put(de.getKey(), de);
			// FG-DE Schlssel in die Tabelle mit den _eakKeys eintragen.
			_eakKeys[i++] = de.getKey();
		}

		// TBD: Triggerdatenart fr Versand gepufferter DE anmelden !!NICHT bei virtuellen EAK !
		// subcribeTrigger();
	}

	// ------------------------------------------------------------------------
	// GETTER / SETTER METHODEN
	// ------------------------------------------------------------------------

	/**
	 * Gibt das Objekt des aktuellen Eak-Umwandlungsmoduls zurck. Weird zum Zugriff auf De-bergreifende Informationen
	 * (Zeitstempel etc.) bentigt.
	 *
	 * @return Objekt des aktuellen Eak-Umwandlungsmoduls.
	 */
	public EakConversionInterface getConversionModul() {
		return _conversionModul;
	}

	/**
	 * Liefert Liste der angeschlossenen De.
	 *
	 * @return Liste der angeschlossenen De.
	 */
	public List getDeList() {
		return _deList;
	}

	/**
	 * Bestimmt das dem EAK zugeordnete Konfigurationsobjekt vom "typ.eak" aus dem Teilmodell TLS.
	 *
	 * @return Das diesem EAK zugeordnete Konfigurationsobjekt.
	 */
	public ConfigurationObject getEak() {
		return _eak;
	}

	/**
	 * Liefert die Bezeichungstext des Eak.
	 *
	 * @return Bezeichnungstext des Gertes
	 */
	public String getEakDescription() {
		return _eakDescription;
	}

	/**
	 * Liefert einen Hashkey auf Basis der Osi7Adresse, der De-Kanal-Nummer und der Funktionsgruppe als String.
	 *
	 * @return Hashkey[] auf Basis der Osi7Adresse, der De-Kanal-Nummer und der Funktionsgruppe ("Osi7Adresse-DeNummer-FG",
	 *         z.B. "51043-193-4").
	 */
	String[] getKeys() {
		return (_eakKeys);
	}

	/**
	 * Liefert die Osi2Adresse des Gertes.
	 *
	 * @return Osi2Adresse des Gertes
	 */
	public int getOsi2Adress() {
		return _osi2Adress;
	}

	/**
	 * Liefert das bergeordnete Sm-Objekt.
	 *
	 * @return bergeordnetes Sm-Objekt.
	 */
	public Sm getSm() {
		return _sm;
	}

	// ------------------------------------------------------------------------
	// KANONISCHE METHODEN
	// ------------------------------------------------------------------------

	/**
	 * Liefert die String-Darstellung eines Eak:
	 * <pre><code>
	 * Eak (Bezeichnung)               :
	 * Knotennummer                    :
	 * Osi2Adresse                     :
	 * Umsetzungsmodul                 :
	 * Schlsselwerte (HashKeys)       :
	 * Verbindungsstatus               :
	 * </pre></code>
	 * Zustzlich werden alle angeschlossenen De mit ihren Eigenschften ausgegen.
	 *
	 * @return String-Darstellung
	 */
	public String toString() {
		StringBuffer sb = new StringBuffer();
		final String NEWLINE = System.getProperty("line.separator");
		final String INDENT1 = "    ";
		sb.append(INDENT1).append("............................................................................").append(NEWLINE);
		sb.append(INDENT1).append("Eak (Bezeichnung)        : ").append(getEakDescription()).append(NEWLINE);
		sb.append(INDENT1).append("Knotennummer             : ");
		sb.append(getSm().getOsi7AdressDecimal()).append(" (Dez)   ");
		sb.append(getSm().getOsi7AdressHex()).append(" (Hex)   ");
		sb.append(getSm().getOsi7AdressLocCode()).append(" (LocCode)").append(NEWLINE);
		sb.append(INDENT1).append("Osi2Adresse              : ").append(getOsi2Adress()).append(NEWLINE);
		sb.append(INDENT1).append("Umsetzungsmodul          : ").append(_conversionModul).append(NEWLINE);
		sb.append(INDENT1).append("Schlsselwerte (HashKeys): ");
		String[] eakKeys = getKeys();
		if (eakKeys.length > 0) {
			sb.append(eakKeys[0]);
		}
		else {
			sb.append("Keine De an diesen Eak angeschlossen!");
		}
		sb.append(NEWLINE);

		for (int i = 1; i < eakKeys.length; ++i) {
			sb.append("                               ").append(eakKeys[i]).append(NEWLINE);
		}
		Iterator i = _deList.iterator();
		while (i.hasNext()) {
			sb.append(i.next().toString());
		}
		return (sb.toString());
	}

	// ------------------------------------------------------------------------
	// INTERFACE METHODEN
	// ------------------------------------------------------------------------

	// --------------------- Interface ClientReceiverInterface ---------------------

	/**
	 * Aktualisierungsmethode, die nach Empfang eines angemeldeten Datensatzes von den
	 * Datenverteiler-Applikationsfunktionen aufgerufen wird. Diese Methode muss von der Applikation zur Verarbeitung der
	 * empfangenen Datenstze implementiert werden. Angemeldet ist hier nur die Attributgruppe
	 * atg.tlsEinzelTelegrammSenden
	 *
	 * @param results Feld mit den empfangenen Ergebnisdatenstzen.
	 */
	public void update(ResultData results[]) {
		// Abhngig vom Attributwert gespeicherte De-Blcke zum Versenden
		// bergeben oder Puffer lschen.
		for (ResultData result : results) {
			if (result.getData() != null) {
				if (result.getData().getUnscaledValue("Senden").getState().getName().equals("senden")) {
					sendBufferedDeBlocks();
				}
				else if (result.getData().getUnscaledValue("Senden").getState().getName().equals("verwerfen")) {
					clearBufferedDeBlocks();
				}
				else {
					debug.error("Unbekannter Wertezustand bei atg.tlsEinzelTelegrammSenden, Attribut 'Senden' erhalten (senden|verwerfen): " + result.getData().getUnscaledValue("Senden").getState().toString() + "\n an EAK:\n" + this.toString());
				}
			}
		}
	}

	// ------------------------------------------------------------------------
	// SONSTIGE METHODEN
	// ------------------------------------------------------------------------

	/**
	 * Speichert einen zu sendenden DE-Blcke fr dieses EAK. Die gespeicherten De-Blcke werden erst nach externer
	 * Triggerung zu einem/mehreren Einzeltelegramm zusammengestellt.
	 *
	 * @param deBlockToSend Der zu speichernde De-Block
	 */
	public void bufferDeBlock(DeBlockToSend deBlockToSend) {
		_bufferedDeBlocks.add(deBlockToSend);
	}

	/**
	 * Verarbeitet ein EinzelTelegramm. Ermittelt zu jedem De-Block das fr die weitere Verarbeitung zustndige De und
	 * delegiert den De-Block an dieses De. Enthlt das Einzeltelegramm als ERSTEN De-Block einen De-Block fr die
	 * DeKanalNummer 255, so werden die Informationen dieses Blocks durch das Eak ausgewertet (Zeitinformationen und
	 * Intervalldaten etc.). Die fr das Einzeltelegramm ermittelten allgemeinen Information stehen dann in den
	 * Umsetzungsmodulen fr die De allen De's des Einzeltelegramms zur Verfgung.
	 *
	 * @param singleTelegram Das zu verarbeitende Einzeltelegramm {@link SingleTelegram}.
	 */
	void processSingleTelegram(SingleTelegram singleTelegram) {
		// Zuerst spezielle Verarbeitung entsprechend dem geladenen Umsetzungsmodul durchfhren...
		debug.finer("Verarbeite Einzeltelegramm: " + Debug.NEWLINE + singleTelegram.toString());
		_conversionModul.processSingleTelegram(singleTelegram);

		// ...danach DeBlcke an die nchste Verarbeitungsstufe (De) weitergeben.
		for (int i = 0; i < singleTelegram.getCountDeBlock(); ++i) {
			DeBlock deBlock = singleTelegram.getDeBlock()[i];

			if (!((deBlock.getDeChannel() == 255) && (i == 0)) || (deBlock.getTyp() == 212) ||
			// NEU wegen Fehler in TLS 2012
			(deBlock.getFg() == 2) && (deBlock.getTyp() == 31)) {

				// Wenn DeKanalNummer == 255 und De-Block ist der ERSTE im Telegramm,
				// dann erfolgte Verarbeitung bereits durch den EAK
				debug.fine("Folgenden De-Block an De zur Weiterverarbeitung delegiert:" + Debug.NEWLINE + deBlock.toString());
				if (_deHashTable.get(deBlock.getKey()) == null) {
					debug.warning(String.format("DeBlock kann nicht verarbeitet werden, da DE/FG an diesem EAK unbekannt (Osi7-De-Fg: %s)%nEAK: %s%n%s", deBlock.getDescriptionKey(), this.toString(), deBlock.toString()));
				}
				else {
					//Telegramm zur Verarbeitung an verantwortlichen DE delegieren
					((De) _deHashTable.get(deBlock.getKey())).processDeBlock(deBlock);
				}
			}
		}
	}

	public void sendBufferedDeBlocks() {
		if (getBufferedDeBlocks().length > 0) {
			TlsApplicationLayer._deBlockQueue.push(getBufferedDeBlocks());
			clearBufferedDeBlocks();
			debug.finer("De-Blocks nach Trigger versandt");
		}
	}

	/**
	 * Liefert die gespeicherten DE-Blcke fr dieses EAK. Die gespeicherten De-Blcke werden erst nach externer Triggerung
	 * zu einem/mehreren Einzeltelegramm zusammengestellt.
	 *
	 * @return Die Liste mit den gespeicherten De-Blcken.
	 */
	public DeBlockToSend[] getBufferedDeBlocks() {
		DeBlockToSend[] t = new DeBlockToSend[_bufferedDeBlocks.size()];
		return _bufferedDeBlocks.toArray(t);
	}

	/** Lscht die gespeicherten DE-Blcke fr dieses EAK. */
	public void clearBufferedDeBlocks() {
		_bufferedDeBlocks.clear();
	}
}
